import re
from typing import List

import pandas as pd

from mindsdb_sql_parser import ast

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.utilities.sql_utils import (
    extract_comparison_conditions, FilterCondition, FilterOperator, SortColumn)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class GithubIssuesTable(APIResource):
    """The GitHub Issue Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository issues" API

        Returns
        -------
        pd.DataFrame
            GitHub issues matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        if limit is None:
            limit = 20

        issues_kwargs = {'state': 'all'}

        if sort is not None:
            for col in sort:
                if col.column in ('created', 'updated', 'comments'):
                    issues_kwargs['sort'] = col.column
                    issues_kwargs['direction'] = 'asc' if col.ascending else 'desc'

                    # supported only 1 column
                    break

        for condition in conditions:
            if (condition.column in ('state', 'assignee', 'creator', 'mentioned', 'milestone')
                    and condition.op == FilterOperator.EQUAL):

                issues_kwargs[condition.column] = condition.value
                condition.applied = True

            elif condition.column == 'labels':
                if condition.op == FilterOperator.IN:
                    issues_kwargs['labels'] = condition.value
                elif condition.op == FilterOperator.EQUAL:
                    issues_kwargs['labels'] = condition.value.split(",")
                condition.applied = True

            elif condition.column == 'updated' and condition.op == FilterOperator.GREATER_THAN:
                issues_kwargs['since'] = condition.value
                condition.applied = True

        self.handler.connect()

        data = []
        count = 0
        for an_issue in self.handler.connection\
                .get_repo(self.handler.repository) \
                .get_issues(**issues_kwargs):
            item = {
                "number": an_issue.number,
                "title": an_issue.title,
                "state": an_issue.state,
                "creator": an_issue.user.login,
                "labels": ",".join(
                    [label.name for label in an_issue.labels]
                ),
                "assignees": ",".join(
                    [
                        assignee.login
                        for assignee in an_issue.assignees
                    ]
                ),
                "comments": an_issue.comments,
                "body": an_issue.body,
                "created": an_issue.created_at,
                "updated": an_issue.updated_at,
                "closed": an_issue.closed_at,
            }

            if 'closed_by' in targets:
                item['closed_by'] = an_issue.closed_by.login if an_issue.closed_by else None

            data.append(item)

            count += 1
            if limit <= count:
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def add(self, issues: List[dict]):
        """Inserts data into the GitHub "Create an issue" API

        Parameters
        ----------
        issues : list of dict

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        if self.handler.connection_data.get("api_key", None) is None:
            raise ValueError(
                "Need an authenticated connection in order to insert a GitHub issue"
            )

        self.handler.connect()
        for issue in issues:
            self._add(issue)

    def _add(self, issue: dict):
        current_repo = self.handler.connection.get_repo(self.handler.repository)

        insert_kwargs = {}

        if issue.get("title", None) is None:
            raise ValueError("Title parameter is required to insert a GitHub issue")

        if issue.get("body", None):
            insert_kwargs["body"] = issue["body"]

        if issue.get("assignees", None):
            insert_kwargs["assignees"] = []
            for an_assignee in issue["assignees"].split(","):
                an_assignee = an_assignee.replace(" ", "")
                try:
                    github_user = self.handler.connection.get_user(an_assignee)
                except Exception as e:
                    raise ValueError(
                        f'Encountered an exception looking up assignee "{an_assignee}" in GitHub: '
                        f"{type(e).__name__} - {e}"
                    )

                insert_kwargs["assignees"].append(github_user)

        if issue.get("milestone", None):
            current_milestones = current_repo.get_milestones()

            found_existing_milestone = False
            for a_milestone in current_milestones:
                if a_milestone.title == issue["milestone"]:
                    insert_kwargs["milestone"] = a_milestone
                    found_existing_milestone = True
                    break

            if not found_existing_milestone:
                logger.debug(
                    f"Milestone \"{issue['milestone']}\" not found, creating it"
                )
                insert_kwargs["milestone"] = current_repo.create_milestone(
                    issue["milestone"]
                )
            else:
                logger.debug(f"Milestone \"{issue['milestone']}\" already exists")

        if issue.get("labels", None):
            insert_kwargs["labels"] = []

            inserted_labels = []
            for a_label in issue["labels"].split(","):
                a_label = a_label.replace(" ", "")
                inserted_labels.append(a_label)

            existing_labels = current_repo.get_labels()

            existing_labels_set = set([label.name for label in existing_labels])

            if not set(inserted_labels).issubset(existing_labels_set):
                new_inserted_labels = set(inserted_labels).difference(
                    existing_labels_set
                )
                logger.debug(
                    "Inserting new labels: " + ", ".join(new_inserted_labels)
                )
                for a_new_label in new_inserted_labels:
                    current_repo.create_label(a_new_label, "000000")

            for a_label in existing_labels:
                if a_label.name in inserted_labels:
                    insert_kwargs["labels"].append(a_label)

        try:
            current_repo.create_issue(issue["title"], **insert_kwargs)
        except Exception as e:
            raise ValueError(
                f"Encountered an exception creating an issue in GitHub: "
                f"{type(e).__name__} - {e}"
            )

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "number",
            "title",
            "state",
            "creator",
            "closed_by",
            "labels",
            "assignees",
            "comments",
            "body",
            "created",
            "updated",
            "closed",
        ]


class GithubPullRequestsTable(APIResource):
    """The GitHub Issue Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:


        """Pulls data from the GitHub "List repository pull requests" API

        Native filters:
        - state: open, closed, or all (default)
        - head: Filter pulls by head user or head organization and branch name
                       in the format of user:ref-name or organization:ref-name
        - base: Filter pulls by base branch name

        Native sorts:
        - created, updated, popularity

        Returns
        -------
        pd.DataFrame
            GitHub pull requests matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        if limit is None:
            limit = 20

        issues_kwargs = {'state': 'all'}

        if sort is not None:
            for col in sort:
                if col.column in ('created', 'updated', 'popularity'):
                    issues_kwargs['sort'] = col.column
                    issues_kwargs['direction'] = 'asc' if col.ascending else 'desc'

                    # supported only 1 column
                    break

        for condition in conditions:
            if (condition.column in ('state', 'head', 'base')
                    and condition.op == FilterOperator.EQUAL):

                issues_kwargs[condition.column] = condition.value
                condition.applied = True

        self.handler.connect()

        data = []
        count = 0
        for a_pull in self.handler.connection\
                .get_repo(self.handler.repository) \
                .get_pulls(**issues_kwargs):

            item = {
                "number": a_pull.number,
                "title": a_pull.title,
                "state": a_pull.state,
                "creator": a_pull.user.login,
                "labels": ",".join(
                    [label.name for label in a_pull.labels]
                ),
                "milestone": a_pull.milestone.title if a_pull.milestone else None,
                "assignees": ",".join(
                    [
                        assignee.login
                        for assignee in a_pull.assignees
                    ]
                ),
                "reviewers": ",".join(
                    [
                        reviewer.login
                        for reviewer in a_pull.requested_reviewers
                    ]
                ),
                "teams": ",".join(
                    [
                        team.name
                        for team in a_pull.requested_teams
                    ]
                ),
                "draft": a_pull.draft,
                "body": a_pull.body,
                "base": a_pull.base.ref if a_pull.base else None,
                "head": a_pull.head.ref if a_pull.head else None,
                "created": a_pull.created_at,
                "updated": a_pull.updated_at,
                "merged": a_pull.merged_at,
                "closed": a_pull.closed_at,
            }

            # downloaded columns, use them only if explicitly requested
            for field in ('comments', 'review_comments', 'mergeable', 'mergeable_state', 'rebaseable',
                          'commits', 'additions', 'deletions', 'changed_files'):
                if field in targets:
                    item[field] = getattr(a_pull, field)
            if 'is_merged' in targets:
                item['is_merged'] = a_pull.merged
            if 'merged_by' in targets:
                item['is_merged'] = a_pull.merged_by.login if a_pull.merged_by else None

            data.append(item)
            count += 1
            if limit <= count:
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "number",
            "title",
            "state",
            "creator",
            "labels",
            "milestone",
            "assignees",
            "reviewers",
            "teams",
            "comments",
            "review_comments",
            "draft",
            "is_merged",
            "mergeable",
            "mergeable_state",
            "merged_by",
            "rebaseable",
            "body",
            "base",
            "head",
            "commits",
            "additions",
            "deletions",
            "changed_files",
            "created",
            "updated",
            "merged",
            "closed",
        ]


class GithubCommitsTable(APIResource):
    """The GitHub Commits Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List commits" API

        Returns
        -------
        pd.DataFrame
            GitHub commits matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        commits_kwargs = {}

        if sort is not None:
            for col in sort:
                if col.column in ("author", "date", "message"):
                    commits_kwargs['sort'] = col.column
                    commits_kwargs['direction'] = 'asc' if col.ascending else 'desc'

                    # supported only 1 column
                    break

        for condition in conditions:
            if condition.column == 'author':
                if condition.op != FilterOperator.EQUAL:
                    raise ValueError("Unsupported where operation for author")
                commits_kwargs["author"] = condition.value
                condition.applied = True

        self.handler.connect()

        data = []
        for a_commit in self.handler.connection.get_repo(
                self.handler.repository
        ).get_commits(**commits_kwargs):

            item = {
                "sha": a_commit.sha,
                "author": a_commit.commit.author.name,
                "date": a_commit.commit.author.date,
                "message": a_commit.commit.message,
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return ["sha", "author", "date", "message"]


class GithubReleasesTable(APIResource):
    """The GitHub Releases Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List repository releases" API

        Returns
        -------
        pd.DataFrame
            GitHub releases matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        self.handler.connect()

        data = []
        for a_release in self.handler.connection.get_repo(
                self.handler.repository
        ).get_releases():

            item = {
                "id": self.check_none(a_release.id),
                "author": self.check_none(a_release.author.login),
                "body": self.check_none(a_release.body),
                "created_at": self.check_none(str(a_release.created_at)),
                "html_url": self.check_none(a_release.html_url),
                "published_at": self.check_none(str(a_release.published_at)),
                "tag_name": self.check_none(a_release.tag_name),
                "title": self.check_none(a_release.title),
                "url": self.check_none(a_release.url),
                "zipball_url": self.check_none(a_release.zipball_url)
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "author",
            "body",
            "created_at",
            "html_url",
            "published_at",
            "tag_name",
            "title",
            "url",
            "zipball_url"
        ]


class GithubBranchesTable(APIResource):
    """The GitHub Branches Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List repository branches" API

        Returns
        -------
        pd.DataFrame
            GitHub branches matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        self.handler.connect()

        data = []
        for branch in self.handler.connection.get_repo(self.handler.repository).get_branches():
            raw_data = branch.raw_data

            item = {
                "name": self.check_none(raw_data["name"]),
                "url": "https://github.com/" + self.handler.repository + "/tree/" + raw_data["name"],
                "commit_sha": self.check_none(raw_data["commit"]["sha"]),
                "commit_url": self.check_none(raw_data["commit"]["url"]),
                "protected": self.check_none(raw_data["protected"])
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "name",
            "url",
            "commit_sha",
            "commit_url",
            "protected"
        ]


class GithubContributorsTable(APIResource):
    """The GitHub Contributors Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List repository contributors" API

        Returns
        -------
        pd.DataFrame
            GitHub contributors matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        self.handler.connect()

        data = []
        for contributor in self.handler.connection.get_repo(self.handler.repository).get_contributors():
            raw_data = contributor.raw_data

            item = {
                "avatar_url": self.check_none(raw_data["avatar_url"]),
                "html_url": self.check_none(raw_data["html_url"]),
                "followers_url": self.check_none(raw_data["followers_url"]),
                "subscriptions_url": self.check_none(raw_data["subscriptions_url"]),
                "organizations_url": self.check_none(raw_data["organizations_url"]),
                "repos_url": self.check_none(raw_data["repos_url"]),
                "events_url": self.check_none(raw_data["events_url"]),
                "received_events_url": self.check_none(raw_data["received_events_url"]),
                "site_admin": self.check_none(raw_data["site_admin"]),
                "name": self.check_none(raw_data["name"]),
                "company": self.check_none(raw_data["company"]),
                "blog": self.check_none(raw_data["blog"]),
                "location": self.check_none(raw_data["location"]),
                "email": self.check_none(raw_data["email"]),
                "hireable": self.check_none(raw_data["hireable"]),
                "bio": self.check_none(raw_data["bio"]),
                "twitter_username": self.check_none(raw_data["twitter_username"]),
                "public_repos": self.check_none(raw_data["public_repos"]),
                "public_gists": self.check_none(raw_data["public_repos"]),
                "followers": self.check_none(raw_data["followers"]),
                "following": self.check_none(raw_data["following"]),
                "created_at": self.check_none(raw_data["created_at"]),
                "updated_at": self.check_none(raw_data["updated_at"])
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "avatar_url",
            "html_url",
            "followers_url",
            "subscriptions_url",
            "organizations_url",
            "repos_url",
            "events_url",
            "received_events_url",
            "site_admin",
            "name",
            "company",
            "blog",
            "location",
            "email",
            "hireable",
            "bio",
            "twitter_username",
            "public_repos",
            "public_gists",
            "followers",
            "following",
            "created_at",
            "updated_at"
        ]


class GithubProjectsTable(APIResource):
    """The GitHub Projects Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List repository projects" API

        Returns
        -------
        pd.DataFrame
            GitHub projects matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        self.handler.connect()

        data = []
        for project in self.handler.connection.get_repo(self.handler.repository).get_projects():
            raw_data = project.raw_data

            item = {
                "owner_url": self.check_none(raw_data["owner_url"]),
                "url": self.check_none(raw_data["url"]),
                "html_url": self.check_none(raw_data["html_url"]),
                "columns_url": self.check_none(raw_data["columns_url"]),
                "id": self.check_none(raw_data["id"]),
                "node_id": self.check_none(raw_data["node_id"]),
                "name": self.check_none(raw_data["name"]),
                "body": self.check_none(raw_data["body"]),
                "number": self.check_none(raw_data["number"]),
                "state": self.check_none(raw_data["state"]),
                "created_at": self.check_none(raw_data["created_at"]),
                "updated_at": self.check_none(raw_data["updated_at"]),
                "creator_login": self.check_none(raw_data["creator"]["login"]),
                "creator_id": self.check_none(raw_data["creator"]["id"]),
                "creator_url": self.check_none(raw_data["creator"]["url"]),
                "creator_html_url": self.check_none(raw_data["creator"]["html_url"]),
                "creator_site_admin": self.check_none(raw_data["creator"]["site_admin"])
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "owner_url",
            "url",
            "html_url",
            "columns_url",
            "id",
            "node_id",
            "name",
            "body",
            "number",
            "state",
            "created_at",
            "updated_at",
            "creator_login",
            "creator_id",
            "creator_url",
            "creator_html_url",
            "creator_site_admin"
        ]

class GithubMilestonesTable(APIResource):
    """The GitHub Milestones Table implementation"""

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        """Pulls data from the GitHub "List repository milestones" API

        Returns
        -------
        pd.DataFrame
            GitHub milestones matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        limit = limit or 20

        self.handler.connect()

        data = []
        for milestone in self.handler.connection.get_repo(self.handler.repository).get_milestones():
            raw_data = milestone.raw_data

            item = {
                "url": self.check_none(raw_data["url"]),
                "html_url": self.check_none(raw_data["html_url"]),
                "labels_url": self.check_none(raw_data["labels_url"]),
                "id": self.check_none(raw_data["id"]),
                "node_id": self.check_none(raw_data["node_id"]),
                "number": self.check_none(raw_data["number"]),
                "title": self.check_none(raw_data["title"]),
                "description": self.check_none(raw_data["description"]),
                "creator": self.check_none(raw_data["creator"]),
                "open_issues": self.check_none(raw_data["open_issues"]),
                "closed_issues": self.check_none(raw_data["closed_issues"]),
                "state": self.check_none(raw_data["state"]),
                "created_at": self.check_none(raw_data["created_at"]),
                "updated_at": self.check_none(raw_data["updated_at"]),
                "due_on": self.check_none(raw_data["due_on"]),
                "closed_at": self.check_none(raw_data["closed_at"])
            }

            data.append(item)

            if limit <= len(data):
                break

        return pd.DataFrame(data, columns=self.get_columns())

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "url",
            "html_url",
            "labels_url",
            "id",
            "node_id",
            "number",
            "title",
            "description",
            "creator",
            "open_issues",
            "closed_issues",
            "state",
            "created_at",
            "updated_at",
            "due_on",
            "closed_at"
        ]


class GithubFilesTable(APIResource):

    def get_path(self, repo, path, file_matches=None, file_not_matches=None, limit=None):

        res = []
        for item in list(repo.get_contents(path)):
            if item.type == "dir":
                subres = self.get_path(repo, item.path, file_matches, file_not_matches, limit)
                res.extend(subres)
                limit -= len(subres)
            else:
                if (
                      (
                        file_matches is None
                        or
                        any(re.match(pattern, item.name) for pattern in file_matches)
                      )
                    and
                      (
                          file_not_matches is None
                          or
                          not any(re.match(pattern, item.name) for pattern in file_not_matches)
                      )
                ):

                    file = {
                        'path': item.path,
                        'name': item.name,
                        'content': item.decoded_content,
                    }
                    res.append(file)
                    limit -= 1
            if limit <= 0:
                break
        return res

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None) -> pd.DataFrame:

        self.handler.connect()
        repo = self.handler.connection.get_repo(self.handler.repository)

        # TODO sort

        path = ''
        file_matches = []
        file_not_matches = []
        for condition in conditions:
            if condition.column == 'path' and condition.op == FilterOperator.EQUAL:
                path = condition.value
                condition.applied = True
            elif condition.column == 'name':
                pattern = f'^{condition.value}$'
                if condition.op == FilterOperator.EQUAL:
                    file_matches.append(pattern)
                elif condition.op == FilterOperator.LIKE:
                    # https://stackoverflow.com/a/26148730
                    pattern = pattern.replace('%', '.*')
                    file_matches.append(pattern)
                elif condition.op == FilterOperator.NOT_LIKE:
                    pattern = pattern.replace('%', '.*')
                    file_not_matches.append(pattern)
                condition.applied = True

        if limit is None:
            limit = 10

        if len(file_matches) == 0:
            file_matches = None
        if len(file_not_matches) == 0:
            file_not_matches = None
        res = self.get_path(repo, path, file_matches, file_not_matches, limit)
        return pd.DataFrame(res, columns=self.get_columns())

    def get_columns(self) -> list:
        return ['path', 'name', 'content']


import requests
from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine


class RayServeHandler(BaseMLEngine):
    """
    The Ray Serve integration engine needs to have a working connection to Ray Serve. For this:
        - A Ray Serve server should be running

    Example:
        
    """  # noqa
    name = 'ray_serve'

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if not args.get('using'):
            raise Exception("Error: This engine requires some parameters via the 'using' clause. Please refer to the documentation of the Ray Serve handler and try again.")  # noqa
        if not args['using'].get('train_url'):
            raise Exception("Error: Please provide a URL for the training endpoint.")
        if not args['using'].get('predict_url'):
            raise Exception("Error: Please provide a URL for the prediction endpoint.")

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        # TODO: use join_learn_process to notify users when ray has finished the training process
        args = args['using']  # ignore the rest of the problem definition
        args['target'] = target
        self.model_storage.json_set('args', args)
        try:
            resp = requests.post(args['train_url'],
                                 json={'df': df.to_json(orient='records'), 'target': target},
                                 headers={'content-type': 'application/json; format=pandas-records'})
        except requests.exceptions.InvalidSchema:
            raise Exception("Error: The URL provided for the training endpoint is invalid.")

        resp = resp.json()
        if resp['status'] != 'ok':
            raise Exception("Error: Training failed: " + resp['status'])

    def predict(self, df, args=None):
        args = self.model_storage.json_get('args')  # override any incoming args for now
        resp = requests.post(args['predict_url'],
                             json={'df': df.to_json(orient='records')},
                             headers={'content-type': 'application/json; format=pandas-records'})
        response = resp.json()

        target = args['target']
        if target != 'prediction':
            # rename prediction to target
            response[target] = response.pop('prediction')

        predictions = pd.DataFrame(response)
        return predictions

    def describe(self, key: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        description = {
            'TRAIN_URL': [args['train_url']],
            'PREDICT_URL': [args['predict_url']],
            'TARGET': [args['target']],
        }
        return pd.DataFrame.from_dict(description)



import os

from flask import Flask, Response
from prometheus_client import generate_latest, multiprocess, CollectorRegistry

from mindsdb.utilities import log

_CONTENT_TYPE_LATEST = str('text/plain; version=0.0.4; charset=utf-8')
logger = log.getLogger(__name__)


def init_metrics(app: Flask):
    prometheus_dir = os.environ.get('PROMETHEUS_MULTIPROC_DIR', None)
    if prometheus_dir is None:
        logger.info("PROMETHEUS_MULTIPROC_DIR environment variable is not set. Metrics server won't be started.")
        return
    elif not os.path.isdir(prometheus_dir):
        os.makedirs(prometheus_dir)
    # See: https://prometheus.github.io/client_python/multiprocess/
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)

    # It's important that the PROMETHEUS_MULTIPROC_DIR env variable is set, and the dir is empty.
    @app.route('/metrics')
    def metrics():
        return Response(generate_latest(registry), mimetype=_CONTENT_TYPE_LATEST)



import os
import numpy as np
import pandas as pd

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.handlers.postgres_handler.postgres_handler import PostgresHandler

logger = log.getLogger(__name__)
os.environ["PGCLIENTENCODING"] = "utf-8"


class RedshiftHandler(PostgresHandler):
    """
    This handler handles connection and execution of the Redshift statements.
    """

    name = 'redshift'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance.
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name, **kwargs)

    def insert(self, table_name: str, df: pd.DataFrame):
        """
        Handles the execution of INSERT statements.

        Args:
            table_name (str): name of the table to insert the data into.
            df (pd.DataFrame): data to be inserted into the table.
        """
        need_to_close = not self.is_connected

        connection = self.connect()

        # Replace NaN values with None
        df = df.replace({np.nan: None})

        # Build the query to insert the data
        columns = ', '.join([f'"{col}"' if ' ' in col else col for col in df.columns])
        values = ', '.join(['%s' for _ in range(len(df.columns))])
        query = f'INSERT INTO {table_name} ({columns}) VALUES ({values})'

        with connection.cursor() as cur:
            try:
                cur.executemany(query, df.values.tolist())
                response = Response(RESPONSE_TYPE.OK)

                connection.commit()
            except Exception as e:
                logger.error(f"Error inserting data into {table_name}, {e}!")
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )

        if need_to_close:
            self.disconnect()

        return response



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

# https://dev.mysql.com/doc/internals/en/connection-phase-packets.html#packet-Protocol::HandshakeResponse

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.classes.client_capabilities import ClentCapabilities
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import CAPABILITIES
from mindsdb.api.mysql.mysql_proxy.classes.server_capabilities import server_capabilities


class HandshakeResponsePacket(Packet):
    '''
    Implementation based on description:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def setup(self, length=0, count_header=1, body=''):
        length = len(body)

        if length == 0:
            return

        self.salt = self.proxy.salt

        self._length = length
        self._seq = count_header
        self._body = body

        self.capabilities = Datum('int<4>')
        self.max_packet_size = Datum('int<4>')
        self.reserved = Datum('string<23>')
        self.username = Datum('string<NUL>')

        self.enc_password = Datum('string<NUL>')
        self.database = Datum('string<NUL>')

        self.charset = Datum('int<1>')

        self.client_auth_plugin = Datum('string<NUL>')

        buffer = body

        if len(body) == 32 and body[9:] == (b'\x00' * 23):
            self.type = 'SSLRequest'
            buffer = self.capabilities.setFromBuff(buffer)
            buffer = self.max_packet_size.setFromBuff(buffer)
            buffer = self.charset.setFromBuff(buffer)
        else:
            self.type = 'HandshakeResponse'
            buffer = self.capabilities.setFromBuff(buffer)
            capabilities = ClentCapabilities(self.capabilities.value)
            buffer = self.max_packet_size.setFromBuff(buffer)
            buffer = self.charset.setFromBuff(buffer)
            buffer = self.reserved.setFromBuff(buffer)
            buffer = self.username.setFromBuff(buffer)

            if server_capabilities.has(CAPABILITIES.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA) \
                    and capabilities.PLUGIN_AUTH_LENENC_CLIENT_DATA:
                self.enc_password = Datum('string<lenenc>')
                buffer = self.enc_password.setFromBuff(buffer)
            elif server_capabilities.has(CAPABILITIES.CLIENT_SECURE_CONNECTION) \
                    and capabilities.SECURE_CONNECTION:
                self.auth_resp_len = Datum('int<1>')
                buffer = self.auth_resp_len.setFromBuff(buffer)
                self.enc_password = Datum(f'string<{self.auth_resp_len.value}>')
                buffer = self.enc_password.setFromBuff(buffer)
            else:
                pass_byte = Datum('int<1>')
                buffer = pass_byte.setFromBuff(buffer)

            if capabilities.CONNECT_WITH_DB:
                buffer = self.database.setFromBuff(buffer)
            if capabilities.PLUGIN_AUTH:
                buffer = self.client_auth_plugin.setFromBuff(buffer)

            # at the end is CLIENT_CONNECT_ATTRS, but we dont use it and dont parse

        self.session.username = self.username.value

    def __str__(self):
        return str({
            'header': {'length': self.length, 'seq': self.seq},
            'username': self.username.value,
            'password': self.enc_password.value,
            'database': self.database.value
        })



import os
import sys
import subprocess
from typing import Text, List


def install_dependencies(dependencies: List[Text]) -> dict:
    """
    Installs the dependencies for a handler by calling the `pip install` command via subprocess.

    Args:
        dependencies (List[Text]): List of dependencies for the handler.

    Returns:
        dict: A dictionary containing the success status and an error message if an error occurs.
    """
    outs = b''
    errs = b''
    result = {
        'success': False,
        'error_message': None
    }
    code = None

    try:
        # Split the dependencies by parsing the contents of the requirements.txt file.
        split_dependencies = parse_dependencies(dependencies)
    except FileNotFoundError as file_not_found_error:
        result['error_message'] = f"Error parsing dependencies, file not found: {str(file_not_found_error)}"
        return result
    except Exception as unknown_error:
        result['error_message'] = f"Unknown error parsing dependencies: {str(unknown_error)}"
        return result

    try:
        # Install the dependencies using the `pip install` command.
        sp = subprocess.Popen(
            [sys.executable, '-m', 'pip', 'install', *split_dependencies],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        code = sp.wait()
        outs, errs = sp.communicate(timeout=1)
    except subprocess.TimeoutExpired as timeout_error:
        sp.kill()
        result['error_message'] = f"Timeout error while installing dependencies: {str(timeout_error)}"
        return result
    except Exception as unknown_error:
        result['error_message'] = f"Unknown error while installing dependencies: {str(unknown_error)}"
        return result

    # Return the result of the installation if successful, otherwise return an error message.
    if code != 0:
        output = ''
        if isinstance(outs, bytes) and len(outs) > 0:
            output = output + 'Output: ' + outs.decode()
        if isinstance(errs, bytes) and len(errs) > 0:
            if len(output) > 0:
                output = output + '\n'
            output = output + 'Errors: ' + errs.decode()
        result['error_message'] = output
    else:
        result['success'] = True

    return result


def parse_dependencies(dependencies: List[Text]) -> List[Text]:
    """
    Recursively parses dependencies from a list of dependencies given in a requirements.txt file for a handler.
    This function will perform the following:
    1. Ignore standalone comments.
    2. Remove inline comments.
    3. Check if the dependency is a path to a requirements file and recursively parse the dependencies from that file.

    Args:
        dependencies (List[Text]): List of dependencies for a handler as read from the requirements.txt file.

    Returns:
        List[Text]: List of parsed dependencies for the handler.
    """
    # get the path to this script
    script_path = os.path.dirname(os.path.realpath(__file__))

    split_dependencies = []
    for dependency in dependencies:
        # ignore standalone comments
        if dependency.startswith('#'):
            continue

        # remove inline comments
        if '#' in dependency:
            dependency = dependency.split('#')[0].strip()

        # check if the dependency is a path to a requirements file
        if dependency.startswith('-r'):
            # get the path to the requirements file
            req_path = dependency.split(' ')[1]
            # create the absolute path to the requirements file
            abs_req_path = os.path.abspath(os.path.join(script_path, req_path.replace('mindsdb/integrations', '..')))
            # check if the file exists
            if os.path.exists(abs_req_path):
                inner_dependencies, inner_split_dependencies = [], []
                # read the dependencies from the file
                inner_dependencies = read_dependencies(abs_req_path)
                # recursively split the dependencies
                inner_split_dependencies = parse_dependencies(inner_dependencies)
                # add the inner dependencies to the split dependencies
                split_dependencies.extend(inner_split_dependencies)
            else:
                raise FileNotFoundError(f"Requirements file not found: {req_path}")

        else:
            split_dependencies.append(dependency)

    return split_dependencies


def read_dependencies(path: Text) -> List[Text]:
    """
    Reads the dependencies for a handler from the relevant requirements.txt file and returns them as a list.

    Args:
        path (Text): Path to the requirements.txt file for the handler.

    Returns:
        List[Text]: List of dependencies for the handler.
    """
    dependencies = []
    # read the dependencies from the file
    with open(str(path), 'rt') as f:
        dependencies = [x.strip(' \t\n') for x in f.readlines()]
        dependencies = [x for x in dependencies if len(x) > 0]
    return dependencies



from urllib.parse import urlparse
import socket
import ipaddress


def is_private_url(url: str):
    """
    Raises exception if url is private

    :param url: url to check
    """

    hostname = urlparse(url).hostname
    if not hostname:
        # Unable find hostname in url
        return True
    ip = socket.gethostbyname(hostname)
    return ipaddress.ip_address(ip).is_private


def clear_filename(filename: str) -> str:
    """
    Removes path symbols from filename which could be used for path injection
    :param filename: input filename
    :return: output filename
    """

    if not filename:
        return filename
    badchars = '\\/:*?\"<>|'
    for c in badchars:
        filename = filename.replace(c, '')
    return filename


def validate_urls(urls, allowed_urls):
    """
    Checks if the provided URL(s) is/are from an allowed host.

    This function parses the URL(s) and checks the network location part (netloc)
    against a list of allowed hosts.

    :param urls: The URL(s) to check. Can be a single URL (str) or a list of URLs (list).
    :param allowed_urls: The list of allowed URLs.
    :return bool:  True if the URL(s) is/are from an allowed host, False otherwise.
    """
    allowed_netlocs = [urlparse(allowed_url).netloc for allowed_url in allowed_urls]

    if isinstance(urls, str):
        urls = [urls]

    # Check if all provided URLs are from the allowed sites
    valid = all(urlparse(url).netloc in allowed_netlocs for url in urls)
    return valid



import os

from duckdb.typing import BIGINT, DOUBLE, VARCHAR, BLOB, BOOLEAN
from mindsdb.interfaces.storage.model_fs import HandlerStorage


def python_to_duckdb_type(py_type):
    if py_type == 'int':
        return BIGINT
    elif py_type == 'float':
        return DOUBLE
    elif py_type == 'str':
        return VARCHAR
    elif py_type == 'bool':
        return BOOLEAN
    elif py_type == 'bytes':
        return BLOB
    else:
        # Unknown
        return VARCHAR


# duckdb doesn't like *args
def function_maker(n_args, other_function):
    return [
        lambda: other_function(),
        lambda arg_0: other_function(arg_0),
        lambda arg_0, arg_1: other_function(arg_0, arg_1),
        lambda arg_0, arg_1, arg_2: other_function(arg_0, arg_1, arg_2),
        lambda arg_0, arg_1, arg_2, arg_3: other_function(arg_0, arg_1, arg_2, arg_2),
    ][n_args]


class BYOMFunctionsController:
    """
    User functions based on BYOM handler
    """

    def __init__(self, session):
        self.session = session

        self.byom_engines = None
        self.byom_methods = {}
        self.byom_handlers = {}

        self.callbacks = {}

    def get_engines(self):
        # get all byom engines
        if self.byom_engines is None:
            # first run
            self.byom_engines = []
            for name, info in self.session.integration_controller.get_all().items():
                if info['type'] == 'ml' and info['engine'] == 'byom':
                    if info['connection_data'].get('mode') == 'custom_function':
                        self.byom_engines.append(name)
        return self.byom_engines

    def get_methods(self, engine):
        if engine not in self.byom_methods:
            ml_handler = self.session.integration_controller.get_ml_handler(engine)

            storage = HandlerStorage(ml_handler.integration_id)
            methods = storage.json_get('methods')
            self.byom_methods[engine] = methods
            self.byom_handlers[engine] = ml_handler

        return self.byom_methods[engine]

    def check_function(self, node):
        engine = node.namespace
        if engine not in self.get_engines():
            return

        methods = self.get_methods(engine)

        fnc_name = node.op.lower()
        if fnc_name not in methods:
            # do nothing
            return

        new_name = f'{node.namespace}_{fnc_name}'
        node.op = new_name

        if new_name in self.callbacks:
            # already exists
            return self.callbacks[new_name]

        def callback(*args):
            return self.method_call(engine, fnc_name, args)

        input_types = [
            param['type']
            for param in methods[fnc_name]['input_params']
        ]

        meta = {
            'name': new_name,
            'callback': callback,
            'input_types': input_types,
            'output_type': methods[fnc_name]['output_type']
        }

        self.callbacks[new_name] = meta
        return meta

    def method_call(self, engine, method_name, args):
        return self.byom_handlers[engine].function_call(method_name, args)

    def create_function_set(self):
        return DuckDBFunctions(self)


class FunctionController(BYOMFunctionsController):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def check_function(self, node):
        meta = super().check_function(node)
        if meta is not None:
            return meta

        # builtin function
        if node.op.lower() == 'llm':
            return self.llm_call_function(node)

    def llm_call_function(self, node):
        name = node.op.lower()

        if name in self.callbacks:
            return self.callbacks[name]

        param_prefix = 'LLM_FUNCTION_'
        chat_model_params = {}
        for k, v in os.environ.items():
            if k.startswith(param_prefix):
                param_name = k[len(param_prefix):]
                if param_name == 'MODEL':
                    chat_model_params['model_name'] = v
                else:
                    chat_model_params[param_name.lower()] = v

        if 'provider' not in chat_model_params:
            chat_model_params['provider'] = 'openai'

        if 'api_key' in chat_model_params:
            # move to api_keys dict
            chat_model_params["api_keys"] = {chat_model_params['provider']: chat_model_params['api_key']}

        try:
            from langchain_core.messages import HumanMessage
            from mindsdb.interfaces.agents.langchain_agent import create_chat_model
            llm = create_chat_model(chat_model_params)
        except Exception as e:
            raise RuntimeError(f'Unable to use LLM function, check ENV variables: {e}')

        def callback(question):
            resp = llm([HumanMessage(question)])
            return resp.content

        meta = {
            'name': name,
            'callback': callback,
            'input_types': ['str'],
            'output_type': 'str'
        }
        self.callbacks[name] = meta
        return meta


class DuckDBFunctions:
    def __init__(self, controller):
        self.controller = controller
        self.functions = {}

    def check_function(self, node):

        meta = self.controller.check_function(node)
        if meta is None:
            return

        name = meta['name']

        if name in self.functions:
            return

        input_types = [
            python_to_duckdb_type(param)
            for param in meta['input_types']
        ]

        self.functions[name] = {
            'callback': function_maker(len(input_types), meta['callback']),
            'input': input_types,
            'output': python_to_duckdb_type(meta['output_type'])
        }

    def register(self, connection):
        for name, info in self.functions.items():
            connection.create_function(
                name,
                info['callback'],
                info['input'],
                info['output'],
                null_handling="special"
            )



import os
import unittest
from mindsdb.integrations.handlers.sharepoint_handler import Handler as SharepointHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class SharepointHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "clientId": os.environ.get('CLIENT_ID'),
                "clientSecret": os.environ.get('CLIENT_SECRET'),
                "tenantId": os.environ.get('TENANT_ID'),
            }
        }
        cls.handler = SharepointHandler('test_sharepoint_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_sites_query(self):
        query = "SELECT * FROM test_sharepoint_handler.sites"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_select_lists_query(self):
        query = "SELECT * FROM test_sharepoint_handler.lists"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_select_siteColumns_query(self):
        query = "SELECT * FROM test_sharepoint_handler.siteColumns"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_select_listItems_query(self):
        query = "SELECT * FROM test_sharepoint_handler.listItems"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('test_sharepoint_handler.siteColumns')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import requests


class OilPriceAPIClient:
    def __init__(self, api_key):
        self.api_key = api_key
        self.base_endpoint = "https://api.oilpriceapi.com/v1/prices"
        self.valid_values_by_type = ["spot_price", "daily_average_price"]
        self.valid_values_by_code = ["BRENT_CRUDE_USD", "WTI_USD"]

    def make_request(self, url, params={}):
        headers = {'Content-type': 'application/json'}
        if self.api_key:
            headers['Authorization'] = 'Token ' + self.api_key
        resp = requests.get(url, headers=headers, params=params)
        content = {}
        if resp.status_code == 200:
            content = {'content': resp.json(), 'code': 200}
        else:
            content = {'content': {}, 'code': resp.status_code, 'error': resp.text}
        return content

    def _is_valid_by_type(self, val):
        return val in self.valid_values_by_type

    def _is_valid_by_code(self, val):
        return val in self.valid_values_by_code

    def create_params_dict(self, by_type, by_code):
        params = {}
        if by_type is not None:
            params["by_type"] = by_type
        if by_code is not None:
            params["by_code"] = by_code
        return params

    def get_latest_price(self, by_type=None, by_code=None):
        url = f'{self.base_endpoint}/latest/'
        params = self.create_params_dict(by_type=by_type, by_code=by_code)
        return self.make_request(url, params=params)

    def get_price_past_day(self, by_type=None, by_code=None):
        url = f'{self.base_endpoint}/past_day/'
        params = self.create_params_dict(by_type=by_type, by_code=by_code)
        return self.make_request(url, params=params)



import pandas as pd
from mindsdb_sql_parser import ast
from pandas import DataFrame

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
)
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class VolumesTable(APITable):
    """
    A class for handling the volumes table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Gets all info about the wanted contents of a bookshelf.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the parameters for the request.
        params = {}
        for op, arg1, arg2 in conditions:
            if op != '=':
                raise NotImplementedError
            if arg1 == 'q' or arg1 == 'download' or arg1 == 'langRestrict' \
                    or arg1 == 'printType'\
                    or arg1 == 'source' or arg1 == 'partner' \
                    or arg1 == 'showPreorders' or arg1 == 'startIndex':
                params[arg1] = arg2
            elif arg1 == 'filter':
                if arg2 not in ['ebooks', 'free-ebooks', 'full', 'paid-ebooks', 'partial']:
                    raise NotImplementedError
                params[arg1] = arg2
            elif arg1 == 'libraryRestrict':
                if arg2 not in ['my-library', 'no-restrictions']:
                    raise NotImplementedError
                params[arg1] = arg2
            elif arg1 == 'printType':
                if arg2 not in ['all', 'books', 'magazines']:
                    raise NotImplementedError
                params[arg1] = arg2
            elif arg1 == 'projection':
                if arg2 not in ['lite', 'full']:
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        # Get the order by from the query.
        if query.order_by is not None:
            if query.order_by[0].value == 'newest':
                params['orderBy'] = 'newest'
            elif query.order_by[0].value == 'relevance':
                params['orderBy'] = 'relevance'
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        # Get the volumes from the Google Books API.
        bookshelves = self.handler.\
            call_application_api(method_name='get_volumes', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(bookshelves) == 0:
            bookshelves = pd.DataFrame([], columns=selected_columns)
        else:
            bookshelves.columns = self.get_columns()
            for col in set(bookshelves.columns).difference(set(selected_columns)):
                bookshelves = bookshelves.drop(col, axis=1)
        return bookshelves

    def get_columns(self) -> list:
        """
        Gets the columns of the table.

        Returns:
            list: List of column names.
        """
        return [
            'kind',
            'id',
            'etag',
            'selfLink',
            'volumeInfo',
            'userInfo'
            'saleInfo',
            'accessInfo',
            'searchInfo',
        ]


class BookshelvesTable(APITable):
    """
    A class for handling the bookshelves table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Gets all info about the wanted bookshelves.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the parameters for the request.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'userId' or arg1 == 'source' or arg1 == 'fields':
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            elif arg1 == 'shelf':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['minShelf'] = arg2
                elif op == '<':
                    params['maxShelf'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        # Get the bookshelves from the Google Books API.
        bookshelves = self.handler.\
            call_application_api(method_name='get_bookshelves', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(bookshelves) == 0:
            bookshelves = pd.DataFrame([], columns=selected_columns)
        else:
            bookshelves.columns = self.get_columns()
            for col in set(bookshelves.columns).difference(set(selected_columns)):
                bookshelves = bookshelves.drop(col, axis=1)
        return bookshelves

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'kind',
            'id',
            'selfLink',
            'title',
            'description',
            'access',
            'updated',
            'created',
            'volumeCount',
            'volumesLastUpdated'
        ]



from mindsdb.integrations.libs.const import HANDLER_TYPE

from mindsdb.integrations.handlers.bedrock_handler.__about__ import __version__ as version, __description__ as description
try:
    from .bedrock_handler import AmazonBedrockHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = 'Amazon Bedrock'
name = 'bedrock'
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'import_error', 'icon_path'
]
